
#include "chains.h"

#include <gtk/gtk.h>
#include <string>
#include <stl.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "global.h"
#include "processor.h"


/*************************************************************** DEFINES */

/* define a rule */
typedef struct
{
	char name[128];
	char rule[1024];
	GtkLabel* label;
}
rule_t;

/* map the rule-label pointer onto a rule definition */
typedef map<GtkLabel*,rule_t> ruletree_t;
typedef ruletree_t::iterator RI;

/* define a chain */
typedef struct
{
	char name[128];
	char default_action[16];
	bool system;									/* if true then it cannot be removed */
	ruletree_t rules;
	GtkLabel* label;
	GtkWidget* subtree;
}
chain_t;

/* map a chain-label pointer onto a chain definition */
typedef map<GtkLabel*,chain_t> chaintree_t;
typedef chaintree_t::iterator CI;


/*************************************************************** VERIABLES */

chaintree_t chain_tree;


/*************************************************************** FUNCTIONS */

/*----------------------------------------- chains only */

/*-----------------------------------------
	initialize the chains (the system ones) */
bool chains_init(void)
{
	printf("chains_init\n");

	// first we ant to clear the tree of any bad items...
	gtk_tree_clear_items(GTK_TREE(rule_tree),0,chain_tree.size());
	// also clear the maps...
	chain_tree.clear();
	
	// now add in the static chains...
	chains_add_chain("input","ACCEPT",true,true);
	chains_add_chain("forward","ACCEPT",false,true);
	chains_add_chain("output","ACCEPT",false,true);
	
	return true;
}

/*-----------------------------------------
	return the chain structure for this name */
chain_t* get_chain_by_name( const char *name )
{
	// ok we must follow through each chain to see what was selected...
	for(CI ci=chain_tree.begin();ci!=chain_tree.end();ci++){
		// we have a chain...
		chain_t& chain = ci->second;
		if( strcmp(chain.name,name) == 0 ){
			// this chain was selected 
			printf("  chain '%s' was found\n",chain.name);
			return &chain;
		}
	}
	return NULL;
}

/*-----------------------------------------
	return the rule structure for this name, given the chain */
rule_t* get_rule_by_name( const char *name, chain_t* chain)
{
	// ok we must follow through each rule to see what was selected...
	for(RI ri=(chain->rules).begin();ri!=(chain->rules).end();ri++){
		// we have a chain...
		rule_t& rule = ri->second;
		if( strcmp(rule.name,name) == 0 ){
			// this chain was selected 
			printf("  rule '%s' was found\n",rule.name);
			return &rule;
		}
	}
	return NULL;	
}



/*-----------------------------------------
	add a new chain with given criteria (duplicates are not allowed) */
bool chains_add_chain(const char* name, const char* default_action, bool select, bool system)
{
	printf("chains_add_chain( '%s', '%s', %d, %d )\n",name, default_action, select, system);

	// see if this name is already taken...
	chain_t* chainptr = get_chain_by_name( name );
	if( chainptr ) return false;
	// else we do not have it yet...
	
	// now add it into the rule_tree
	GtkWidget *item, *subtree;
	// main branch add
	item = gtk_tree_item_new_with_label( (gchar*)name );
	gtk_tree_append( GTK_TREE(rule_tree), item );
	// select and show
	if( select ) gtk_tree_select_child( GTK_TREE(rule_tree), item );
	gtk_widget_show( GTK_WIDGET( item ) );
	// make it a subtree
	subtree  = gtk_tree_new();
	gtk_tree_item_set_subtree( GTK_TREE_ITEM( item ), subtree );
	// store for suture subtree additions
	
	// now add it into the tree
	GtkLabel* label = GTK_LABEL(GTK_BIN(item)->child);
	strcpy(chain_tree[label].name,name);
	strcpy(chain_tree[label].default_action,default_action);
	chain_tree[label].system = system;
	chain_tree[label].rules.clear();
	chain_tree[label].label = label;
	chain_tree[label].subtree = subtree;
	
	return true;
}

/* delete a chain given the GtkLabel pointer */
bool chains_delete_chain(const GtkLabel* label)
{
	printf("chains_delete_chain( %p )\n", label);

	// first see if this chain exists or not
	CI ci = chain_tree.find((GtkLabel*)label);
	if( ci==chain_tree.end() ) return false;
	
	// ok, so now we have a chain_t
	chain_t& chain = ci->second;
	
	// see if this is a system chain, if so we cannot remove it
	if( chain.system ) return false;
	
	printf("... not written yet\n");

	// now we must remove all the rules in it...
	// for each
	//   remove a rule from display
	//   remove a rule from map
	
	// now remove from display...
	// gtk_*( GTK_*(rule_tree) )...

	// now remove the chain
	// chain_tree.erase(chain);	
	
	return true;
}

/*-----------------------------------------
 rename a chain */
bool chains_rename_chain(const GtkLabel* label, const char* new_name)
{
	printf("chains_rename_chain( %p, '%s' )\n",label, new_name);

	// first see if this chain exists or not
	CI ci = chain_tree.find((GtkLabel*)label);
	if( ci==chain_tree.end() ) return false;
	
	// ok, so now we have a chain_t
	chain_t& chain = ci->second;
	
	// see if this is a system chain, if so we cannot remove it
	if( chain.system ) return false;
	
	printf("... not written yet\n");
	
	
	return true;
}

/*-----------------------------------------
 change default action */
bool chains_change_default_action(const GtkLabel* label, const char* action)
{
	printf("chains_change_default_action( %p, '%s' )\n",label,action);

	// first see if this chain exists or not
	CI ci = chain_tree.find((GtkLabel*)label);
	if( ci==chain_tree.end() ) return false;
	
	// ok, so now we have a chain_t
	chain_t& chain = ci->second;
	
	if( chain.system )
		printf("... not written yet\n");
	else
		printf("... not written yet\n");


	return true;
}

/*----------------------------------------- rules only */

/*-----------------------------------------
	add a new rule to the end of the chain */
bool chains_append_rule(const GtkLabel* chain_label, const char* rule_name, const char* rule_def, bool select)
{
	printf("chains_append_rule( %p, '%s', '%s' )\n",chain_label,rule_name,rule_def);
	
	// first see if this chain exists or not
	CI ci = chain_tree.find((GtkLabel*)chain_label);
	if( ci==chain_tree.end() ) return false;
	
	// ok, so now we have a chain_t
	chain_t& chain = ci->second;
	
	// second see if this rule exists or not (on this chain)
	rule_t* rule = get_rule_by_name( rule_name, &chain);
	if( rule ) return false;
	
	// now add the rule to the display
	GtkWidget *item, *subtree;
	subtree = chain.subtree;
	// create a new item and add it to the tree
	item = gtk_tree_item_new_with_label( (gchar*)rule_name );
	gtk_tree_append( GTK_TREE(subtree), item );
	// open the branch
	gtk_tree_item_expand( GTK_TREE_ITEM( GTK_TREE(subtree)->tree_owner ) );

	// now we add this rule definition to the chain rule list
	GtkLabel *label = GTK_LABEL(GTK_BIN(item)->child);
	strcpy(chain.rules[label].name, rule_name);
	strcpy(chain.rules[label].rule, rule_def);
	chain.rules[label].label = GTK_LABEL(GTK_BIN(item)->child);

	printf("   rule[%s]='%s'\n",chain.rules[label].name,chain.rules[label].rule);
	

	// now select and show
	if( select ) gtk_tree_select_child( GTK_TREE(rule_tree), item );
	gtk_widget_show( GTK_WIDGET( item ) );

	return true;
}

/*-----------------------------------------
 delete a rule from the chain */
bool chains_delete_rule(const GtkLabel* label, const char* rule_name)
{
	printf("chains_delete_rule( %p, '%s' )\n",label,rule_name);
	return true;
}

/*-----------------------------------------
	rename a rule */
bool chains_rename_rule(const GtkLabel* label, const char* rule_name, const char* new_name)
{
	printf("chains_rename_rule( %p, '%s', '%s' )\n",label,rule_name,new_name);
	return true;
}

/*-----------------------------------------
	change the rule definition */
bool chains_redefine_rule(const GtkLabel* label, const char* rule_name, const char* rule_def)
{
	printf("chains_redefine_rule( %p, '%s', '%s' )\n",label,rule_name,rule_def);
	return true;
}

/*----------------------------------------- integrated */

/*-----------------------------------------
	something was selected */
bool chains_handle_selection(GtkLabel* tree_label)
{
	printf("chains_handle_selection( GtkLabel* )\n");
	
	// ok we must follow through each chain to see what was selected...
	for(CI ci=chain_tree.begin();ci!=chain_tree.end();ci++){
		// we have a chain...
		chain_t& chain = ci->second;
		if( (chain.label) == (tree_label) ){
			// this chain was selected 
			printf("  chain '%s' was selected\n",chain.name);
			return true;
		}
		// we have to see if this chain has any chaildern that may have been selected
		for(RI ri=chain.rules.begin();ri!=chain.rules.end();ri++){
			// we have a rule...
			rule_t& rule = ri->second;
			if( (rule.label) == (tree_label) ){
				// this rule was selected
				printf("  rule '%s' was selected\n",rule.name);
				rule_to_display(rule.name,rule.rule);
				return true;
			}
		}
	}
	return false;
}

/*-----------------------------------------
	load from file descriptor */
bool chains_load_from_file(int fd)
{
	// first clear all info
	if( !chains_init() ) return false;
	
	// open a stream on this file descriptor
	FILE *file = fdopen(fd, "r");
	
	// now start with the reading of the stream
	char buffer[4096],*ptr;
	int len,counter=0;
	
	while( fgets( buffer, 4096, file ) ){
		ptr=buffer;
		len=strlen(ptr);
		// get rid off the trailing carriage return
		while( len>0 && (ptr[len-1]=='\n' || ptr[len-1]=='\r') ) 
			{ptr[--len]=0;}
		// skip any leading spaces
		while( (*ptr)&&isspace(*ptr) ) ptr++;
		// see what kind of an enry we have...
		if( strncmp(ptr,":",1)==0 ){
			// define a chain, get both chain and default action
			char name[128], action[128];
			sscanf(buffer,":%s %s",name,action);
			if( !chains_add_chain(name,action) ){
				chain_t* chain = get_chain_by_name(name);
				if(chain)
					chains_change_default_action(chain->label,action);
			}
		}else if( strncmp(ptr,"-A",2)==0 ){
			// append a rule, get both chain name and rule definition
			char *chain_name, *rule, name[256];
			ptr+=3;										// skip over the '-A '
			chain_name=ptr;
			while( (*ptr)&&(!isspace(*ptr)) ) ptr++;
			*ptr=0;										// null terminate the string
			rule=ptr+1;								// the rest is the rule definition
			// make up a name
			bzero(name,256);
			snprintf(name,255,"rule #%d",++counter);
			chain_t* chain = get_chain_by_name(chain_name);
			if(chain){
				if( !chains_append_rule(chain->label, name, rule) )
					chains_redefine_rule(chain->label, name, rule);
			}
		}
	}
	
	fclose(file);
	
	return true;
}

/*-----------------------------------------
	save to a file descriptor */
bool chains_save_to_file(int fd)
{	
	// open a stream on this file descriptor
	FILE *file = fdopen(fd, "w");

	
	fclose(file);
	return true;
}


/*-----------------------------------------
	perform the file action */
bool do_file_action( char* filemode, char* filename )
{
	if( strcmp(filemode,"open")==0 ){
		
		printf("===> LOADING FILE '%s'\n", filename);
		
		int fd = open(filename,O_RDONLY);
		if( !fd ) return false;
		
		// do a read
		bool ret = chains_load_from_file(fd);
		
		close(fd);
		return ret;
			
	}else if( strcmp(filemode,"save")==0 ){
		
		printf("===> SAVING TO FILE '%s'\n", filename);
		
		int fd = open(filename,O_WRONLY|O_CREAT,0600);
		if( !fd ) return false;
		
		// do a write
		bool ret = chains_save_to_file(fd);
		
		close(fd);
		return ret;

	}
	
	return false;
}
